#include <avr/io.h>
#include <avr/interrupt.h>

#include <util/delay.h>

#include <stdio.h>
#include <string.h>

#include "pff.h"
#include "usart.h"


FATFS fs;					// sistemul de fisiere

/*---------------------------------------------------------------------------*/
/* Player audio                                                              */
/*---------------------------------------------------------------------------*/

/*
 * Four-Character Code - folosit pentru a indentifica formate de date
 */
#define FCC(c1, c2, c3, c4) \
	(((DWORD)(c4) << 24) + \
	 ((DWORD)(c3) << 16) + \
	 (( WORD)(c2) <<  8) + \
	 (( BYTE)(c1) <<  0))


	 uint8_t	buf[2][256];	// wave output buffers (double buffering)
const	 uint16_t	buf_size = 256;	// front and back buffer sizes
volatile uint8_t	buf_front = 0;	// front buffer index (current buffer used)
volatile uint8_t	buf_pos = 0;	// current buffer position
volatile uint8_t	buf_sync = 0;

#define BUF_FRONT	(buf[buf_front])
#define BUF_BACK	(buf[1 - buf_front])


ISR(TIMER0_COMPA_vect)
{
	OCR1A = BUF_FRONT[buf_pos++];

	// swap buffers when end is reached (end is 256 <=> overflow to 0)
	if(buf_pos == 0)
		buf_front = 1 - buf_front;
}

void timer0_start(void)
{
	// interrupt on compare A
	TIMSK0 |= (1 << OCIE0A);
	// CTC, top OCRA
	TCCR0B |= (0 << WGM02);
	TCCR0A |= (1 << WGM01) | (0 << WGM00);
	// prescaler 8
	TCCR0B |= (2 << CS00);
}

void timer0_stop(void)
{
	TCCR0B = 0;
	TCCR0A = 0;
	TIMSK0 = 0;
	OCR0A = 0;
	TCNT0 = 0;
}

void timer1_start(void)
{
	// 8-bit FastPWM
	TCCR1B |= (0 << WGM13) | (1 << WGM12);
	TCCR1A |= (0 << WGM11) | (1 << WGM10);
	// channel A inverted
	TCCR1A |= (3 << COM1A0);
	// prescaler 1
	TCCR1B |= (1 << CS10);
}

void timer1_stop(void)
{
	TCCR1B = 0;
	TCCR1A = 0;
	OCR1A = 0;
	TCNT1 = 0;
}

BOOL continue_play()
{
	if((PINA & (1 << PA3)) == 0 && (PINA & (1 << PA4)) == 0)
		return FALSE;

	return TRUE;
}


/*
 * Incarca header-ul unui fisier WAVE
 *
 * @return DWORD
 * 	0 => format invalid
 * 	1 => eroare I/O
 * 	>1 => numarul de sample-uri
 */
DWORD load_header(void)
{
	DWORD size;
	WORD ret;

	// citeste header-ul (12 octeti)
	if(pf_read(BUF_FRONT, 12, &ret))
		return 1;

	if(ret != 12 || LD_DWORD(BUF_FRONT + 8) != FCC('W','A','V','E'))
		return 0;

	for(;;)
	{
		// citeste chunk ID si size
		pf_read(BUF_FRONT, 8, &ret);
		if(ret != 8)
			return 0;

		size = LD_DWORD(&BUF_FRONT[4]);

		// verifica FCC
		switch(LD_DWORD(&BUF_FRONT[0]))
		{
			// 'fmt ' chunk
			case FCC('f','m','t',' '):
				// verifica size
				if(size > 100 || size < 16) return 0;

				// citeste continutul
				pf_read(BUF_FRONT, size, &ret);
				// verifica codificarea
				if(ret != size || BUF_FRONT[0] != 1) return 0;
				// verifica numarul de canale
				if(BUF_FRONT[2] != 1 && BUF_FRONT[2] != 2) return 0;
				// verifica rezolutia
				if(BUF_FRONT[14] != 8 && BUF_FRONT[14] != 16) return 0;

				// seteaza sampling rate-ul
				OCR0A = (BYTE)(F_CPU / 8 / LD_WORD(&BUF_FRONT[4])) - 1;
				break;

			// 'data' chunk => incepe redarea
			case FCC('d','a','t','a'):
				return size;

			// 'LIST' chunk => skip
			case FCC('L','I','S','T'):
			// 'fact' chunk => skip
			case FCC('f','a','c','t'):
				pf_lseek(fs.fptr + size);
				break;

			// chunk necunoscut => eroare
			default:
				return 0;
		}
	}

	return 0;
}

/*
 * Functie care reda un fisier audio
 *
 * path - calea absoluta a fisierului
 *
 * @return UINT
 *	FR_OK daca a rulat cu succes fisierul
 */
UINT play(const char *path)
{
	FRESULT ret;

	if((ret = pf_open(path)) == FR_OK)
	{
		WORD bytes_read;

		// incarca header-ul fisierului
		DWORD current_size = load_header();
		if(current_size < buf_size)
			return FR_NO_FILE;

		// align to sector boundary
		ret = pf_lseek((fs.fptr + 511) & ~511);
		if(ret != FR_OK)
			return ret;

		// fill front buffer
		ret = pf_read(BUF_FRONT, buf_size, &bytes_read);
		if(ret != FR_OK)
			return ret;
		if(bytes_read < buf_size)
			return ret;

		// reset front buffer index
		buf_pos = 0;

		// start output
		timer0_start();
		timer1_start();
		DDRD |= (1 << PD5);

		while(continue_play())
		{
			uint8_t old_buf_front = buf_front;
			
			// fill back buffer
			ret = pf_read(BUF_BACK, buf_size, &bytes_read);
			if(ret != FR_OK)
				break;
			if(bytes_read < buf_size)
				break;

			// wait for buffer swap
			while(old_buf_front == buf_front) ;
		}

		// stop output
		DDRD &= ~(1 << PD5);
		timer1_stop();
		timer0_stop();
	}

	return ret;
}

/*---------------------------------------------------------------------------*/
/* Reda melodiile                                                            */
/*---------------------------------------------------------------------------*/

#define MUSIC "Replici"


/*
 * Cauta si returneaza al n-lea fisier cu extensia ".WAV" din folder-ul primit
 */
void get_music(int n, const char *folder, char *filename)
{
	filename[0] = '\0';

	DIR dir;
	if(pf_opendir(&dir, folder) != FR_OK)
		return;

	FILINFO file;
	while(n >= 0 && pf_readdir(&dir, &file) == FR_OK &&
		file.fname != NULL && file.fname[0] != '\0')
	{
		if((file.fattrib & (AM_DIR | AM_HID)) == 0 &&
			strstr(file.fname, ".WAV") != NULL)
		{
			n--;
		}
	}

	strcpy(filename, file.fname);
}

void change_song()
{
	static int track = 0;
	static char path[] = MUSIC"/01-Hippy.WAV";
	char * const filename = &path[sizeof(MUSIC)];

	if(!(PINA & (1 << PA3)) || !(PINA & (1 << PA4)))
	{

		DIR dir;
		if(pf_opendir(&dir, MUSIC) != FR_OK)
		{
			return;
		}
	}

	if(!(PINA & (1 << PA3)) && (PINA & (1 << PA4)))
	{
		track++;

		filename[-1] = '\0';
		get_music(track, path, filename);
		filename[-1] = '/';

		if(filename[0] == '\0')
		{
			track = 0;

			filename[-1] = '\0';
			get_music(track, path, filename);
			filename[-1] = '/';
		}

		_delay_ms(50);
		PORTA ^= (1 << PA2);
		play(path);
	}
	else if(!(PINA & (1 << PA4)))
	{
		
		track--;

		filename[-1] = '\0';
		get_music(track, path, filename);
		filename[-1] = '/';

		if(filename[0] == '\0')
		{
			track = 0;

			filename[-1] = '\0';
			get_music(track, path, filename);
			filename[-1] = '/';
		}

		_delay_ms(50);
		PORTA ^= (1 << PA2);
		play(path);
	}
}

void init()
{
	/* Setez pinii PA4 si PA3 ca intrare. */
	DDRA &= ~((1 << PA3) | (1 << PA4));

	/* Activez rezistenta de pull-up pentru pinii PA4 si PA3. */
	PORTA |= (1 << PA3) | (1 << PA4);

	/* Setez pinul PD5 ca iesire. */
	DDRD |= (1 << PD5);
	

	/* Opresc buzzer-ul. */
	PORTD &= ~(1 << PD5);
	
	/* Setez pinul PA2 ca iesire. */
	DDRA |= (1 << PA2);
	

	/* Opresc LED-ul. */
	PORTA &= ~(1 << PA2);
	
	
	DIR dir;
	if(pf_opendir(&dir, "Replici") != FR_OK)
	{
		return;
	}
	
	FILINFO file;
	while(pf_readdir(&dir, &file) == FR_OK &&
		file.fname != NULL && file.fname[0] != '\0')
	{
		change_song();
		_delay_ms(200);
	}
}


int main (void)
{
	USART0_init();
	
	printf("Hello debug\n");

	sei();

	for(;;)
	{
		// mount filesystem
		if(pf_mount(&fs) != FR_OK)
		{
			// wait a while and retry
			_delay_ms(1000);
			continue;
		}

		for(;;)
		{
			init();

			_delay_ms(200);
		}
	}

	return 0;
}
